comment #

   U7RUNINT.ASM
  --------------

  Code for the interrupt service routines used by U7RUN.

  Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>

  Distributed under the terms of the GNU General Public License.  For more
  information please read the file COPYING which is included in the
  distribution.

#

locals @@

.model small,C

.data

; DWORD alignment
temp   dd ?

mainstack dd ?
tempstack dd ?

; WORD alignment
thisbp dw ?
newsize dw ?

; BYTE alignment
prefix db ?
inaltstack db 0

segfaulttext db 10
	db "Oh no -- something bad just happened.",10,10
	db "Error code:    %04X - %ld",10,10
	db "Original CS:IP %04X:%04X",10
	db "Current  CS:IP %04X:%04X",10
	db "Stack pointer: %04X:%04X",10,10
	db "EAX %08lX  EBX %08lX  ECX %08lX  EDX %08lX",10
	db "ESP %08lX  EBP %08lX  ESI %08lX  EDI %08lX",10
	db "DS  %04X      ES  %04X      FS  %04X      GS  %04X",10,10
	db "Code at original CS:IP was",10
	db 10 dup ("%02X "),"(%02X) ",9 dup ("%02X "),10
	db 10*3 dup (" ")," ^^ CS:IP",10,10
	db "Please send this information, together with a description of what you",10
	db "were doing and your system setup to jdrexler@julian.uwo.ca",10
	db "On Windows 95, press Alt-Enter to switch to window mode, then you can",10
	db "cut&paste this text for easier reporting.",10,10
	db "Press any key to terminate the program with a GPF (may lock up in DOS).",10

extrn intcalls:dword:34,_psp:word

public newsize

.code
.386

public threebyteop,fourbyteop,stringop,movimm,xlatint,special,segfault
public savestack,reducesize

extrn getvoodoomem:proc,setvoodoomem:proc,printf:proc

; selected segment prefix
noprf equ 0 ; (no prefix)
esprf equ 1
csprf equ 2
ssprf equ 3
dsprf equ 4
fsprf equ 5
gsprf equ 6
segmask equ 7	; to mask out the segment prefix
msprf equ 8	; memory size (66h)
osprf equ 16	; operand size (67h); shouldn't be present (replaced by the int call)
lockprf equ 32	; lock (f0h) - shouldn't be present
repnzprf equ 64	; repnz (f2h)
repzprf equ 128	; rep or repz depending on string op (f3h)

; base offsets into bp for accessing the old register values
SEGBASE	equ 6			; the stack frame: cs, ip, bp
SEGSIZE equ 6*2			; six segment registers start at BP
REGBASE equ SEGBASE+SEGSIZE
REGSIZE equ 8*4			; eight regular registers start after that
INTBASE equ REGBASE+REGSIZE
INTSIZE equ 3*2			; and finally cs,ip and the flags

; original values.  These are the values stored by the pushad, in the
; same order as the RRR field in any postbyte, i.e. eax,ecx,edx,ebx,esp,ebp,esi,edi
stackseg = ss	; may be fs when we are in the alternate stack frame
oldeax equ stackseg:[bp+REGBASE+28]
oldebx equ stackseg:[bp+REGBASE+16]
oldecx equ stackseg:[bp+REGBASE+24]
oldedx equ stackseg:[bp+REGBASE+20]
oldesp equ stackseg:[bp+REGBASE+12]
oldebp equ stackseg:[bp+REGBASE+8]
oldesi equ stackseg:[bp+REGBASE+4]
oldedi equ stackseg:[bp+REGBASE+0]

oldes  equ word ptr stackseg:[bp+SEGBASE+10]
thiscs equ word ptr stackseg:[bp+SEGBASE+8]	; this is really the current cs - not the old one
oldss  equ word ptr stackseg:[bp+SEGBASE+6]
oldds  equ word ptr stackseg:[bp+SEGBASE+4]
oldfs  equ word ptr stackseg:[bp+SEGBASE+2]
oldgs  equ word ptr stackseg:[bp+SEGBASE+0]

; these are stored by the int call itself
oldflags equ word ptr stackseg:[bp+INTBASE+4]
oldcs    equ word ptr stackseg:[bp+INTBASE+2]
oldip    equ word ptr stackseg:[bp+INTBASE+0]

dirflag equ 400h	; the bit which contains the direction flag in the flags register

;----------Macros-------------------------

; code at the start of each interrupt handler.  Saves all registers, and
; checks the prefix
; NOTE: if anything is changed here, SEGBASE et.al. have to be modified too!
intro macro
	int 3
	pushad
	push es		; these don't get pushed by pushad
	push cs		; cs and ss only so they can easily be referenced
	push ss
	push ds
	push fs
	push gs
	push cs         ; make a standard stack frame for debugging
	call @@puship	; equivalent to "push ip"
@@puship:
	push bp
	mov bp,sp

	call checkprefix
endm

; code before returning from the handler.  Has a save stack restore,
; which works even if some pushes weren't popped
extro macro
	mov sp,bp
	add sp,6	; remove the fake stack frame
	pop gs
	pop fs
	pop ds
	add sp,4	; don't pop ss or cs... not a good idea to do
	pop es
	popad
	iret
endm

; skips the parameter bytes when returning from the int call by
; adjusting the return IP
adjustip macro howmuch
	ifidn <howmuch>,<1>
		inc oldip
	else
		add oldip,howmuch
	endif
endm

sizemultiplex macro size,sizebyte,sizeword,sizedword
	cmp size,4
	je short sizedword
	cmp size,2
	je short sizeword
endm

; wrappers for calling the voodoo access routines, save all registers
voodoogetmem macro address,memsize
	oldstack
	pushad
	push memsize
	push address
	call getvoodoomem
	mov temp,eax	; eax gets destroyed by popad...
	add sp,6
	popad
	mov eax,temp
	newstack
endm

voodoosetmem macro address,value,memsize
	oldstack
	pushad
	push memsize
	push value
	push address
	call setvoodoomem
	add sp,10
	popad
	newstack
endm

; executes a real CPU instruction
realinstruction macro instr,label
	push oldflags   ; restore the oldflags into the flag register
	popf            ; this is necessary for example for adc etc.
label:
	instr
	pushf           ; and save the new flags
	pop oldflags
endm

; restores the original stack for calling C functions which rely on DS==SS
oldstack macro checkstate
	ifnb <checkstate>
		cmp inaltstack,0
		jnz short @@dontswitch
	endif
	push fs
	mov word ptr [tempstack+2],ss
	mov word ptr tempstack,sp
	cli
	lss sp,mainstack
	sti
	mov fs,word ptr [tempstack+2]
	stackseg = fs
	mov inaltstack,1

	ifnb <checkstate>
		mov inaltstack,0
		jmp short @@switchdone
@@dontswitch:
		inc inaltstack
@@switchdone:
	endif
endm

; and switch back to the current stack
newstack macro checkstate
	ifnb <checkstate>
		sub inaltstack,1
		jnz short @@stillneeded
	endif
	cli
	lss sp,tempstack
	sti
	pop fs
	stackseg = ss
	mov inaltstack,0
	ifnb <checkstate>
@@stillneeded:
	endif
endm

;----------Supplemental code-------------------
;
; See what (if any) prefixes were used, and record them.
; Note: the segment register prefix should be irrelevant; the only selector
;	which has access to voodoo memory is 0000h
checkprefix proc
	if @datasize eq 0
		mov bx,cs
	else
		mov bx,@data
	endif
	mov ds,bx
	mov thisbp,bp
	mov es,oldcs
	mov bx,oldip
	push bx
	sub bx,2		; first byte of the int call

		; gather statistics
	xor ax,ax
	mov al,byte ptr es:[bx+1]
	mov dl,al
	sub al,0a0h
	shl ax,2
	mov di,ax
	inc intcalls[di]
	inc intcalls[33*4]

	mov prefix,0

	;es cs ss ds fs gs MS  OS  LOCK REPNZ REP/REPZ
	;26 2e 36 3e 64 65 66 (67) f0   f2    f3
@@nextprefix:
	dec bx
	mov al,byte ptr es:[bx]
	test prefix,segmask
	jnz short @@testnonsegprefix	; can only have one segment prefix
	cmp al,26h
	jne short @@testcs
	or prefix,esprf
	jmp @@nextprefix
@@testcs:
	cmp al,2eh
	jne short @@testss
	or prefix,csprf
	jmp @@nextprefix
@@testss:
	cmp al,36h
	jne short @@testds
	or prefix,ssprf
	jmp @@nextprefix
@@testds:
	cmp al,3eh
	jne short @@testfs
	or prefix,dsprf
	jmp @@nextprefix
@@testfs:
	cmp al,64h
	jne short @@testgs
	or prefix,fsprf
	jmp @@nextprefix
@@testgs:
	cmp al,65h
	jne short @@testMS
	or prefix,gsprf
	jmp @@nextprefix

@@testnonsegprefix:

@@testMS:
	cmp al,66h
	jne short @@testOS
	or prefix,msprf
	jmp @@nextprefix
@@testOS:
	cmp al,67h
	jne short @@testlock
	or prefix,osprf
	jmp @@nextprefix
@@testlock:
	cmp al,0f0h
	jne short @@testrepnz
	or prefix,lockprf
	jmp @@nextprefix
@@testrepnz:
	cmp al,0f2h
	jne short @@testrepz
	or prefix,repnzprf
	jmp @@nextprefix
@@testrepz:
	cmp al,0f3h
	jne short @@prefixdone
	or prefix,repzprf
	jmp @@nextprefix

@@prefixdone:
	pop bx		; now es:[bx] points to bytes following the int call

	; make sure that the selected segment prefix is a 0000 selector
@@checkagain:
	xor eax,eax
	mov al,prefix
	and al,segmask
	jz short @@needdefaultprefix
	dec ax
	shl ax,1
	neg ax
	mov si,ax
	cmp word ptr oldes[si],0
	je short @@goodsegment

	; bad one, cause a segfault to draw attention
	push 16
	call segfault

@@goodsegment:
	ret

@@needdefaultprefix:
	cmp dl,0b2h+6
	jb short @@nostos
	cmp dl,0b2h+7
	ja short @@nostos
	or prefix,esprf		; stosb and stos have default es
	jmp @@checkagain
@@nostos:
	or prefix,dsprf		; no prefix given. assume it's ds
	jmp @@checkagain

checkprefix endp

; called when something unexpected occurs.  Generates debug output, waits for
; a key and then segfaults.
segfault proc
	int 3
	push bp
	mov bp,sp
	mov di,[bp+4]	; error code
	shl edi,16
	mov di,sp
	add di,2
	push edi
	push bx
	mov bp,thisbp	; restore the original bp for the oldeax variables to work
	mov ax,3
	int 10h		; restore text screen
	pop bx
	pop edi

	; switch to C stack, if it isn't already active
	oldstack 1

	; push the code surrounding the problem

	std		; push in reverse order
	mov cx,20
	mov bx,ds
	mov dx,oldcs
	mov ds,dx
	mov si,oldip
	add si,9
	xor ah,ah

@@nextcodebyte:
	lodsb
	push ax
	loop @@nextcodebyte
	cld
	mov ds,bx

	push oldgs
	push oldfs
	push oldes
	push oldds
	push dword ptr oldedi
	push dword ptr oldesi
	push dword ptr oldebp
	push dword ptr oldesp
	push dword ptr oldedx
	push dword ptr oldecx
	push dword ptr oldebx
	push dword ptr oldeax
	push word ptr oldesp
	push oldss
	push fs:word ptr [di]
	shr edi,16
	push cs
	push oldip
	push oldcs
	push intcalls[32*4]
	push di		; error code
	push offset segfaulttext

	call printf
	add sp,12*2+8*4

	newstack		; definitely restore the stack

	; wait for a key
	mov ax,40h
	mov es,ax
@@keepwaiting:			; a busy loop makes debugging easier because a
	mov ax,es:[1ah]		; debug break will very likely happen here and
	cmp ax,es:[1ch]		; not in some Windows code
	je @@keepwaiting
	mov es:[1ch],ax		; clear the key buffer
	xor eax,eax
	mov eax,[eax-1]	; segfault

	pop bp
	ret
segfault endp

savestack proc
	mov word ptr [mainstack+2],ss
	push bp
	sub bp,127	; safety frame of 127 bytes
	mov word ptr mainstack,bp
	pop bp
	ret
savestack endp

;void reducesize(int howmuch);
reducesize proc
	arg howmuch

	int 3

	mov si,-1
	mov cx,sp
	neg cx		; this many bytes in the stack right now
	mov di,howmuch
	mov ax,di
	neg di		; end of new stack
	dec di		; last byte on new stack
	std		; count downwards like a stack
	mov bx,ds
	mov es,bx
	rep movsb
	cld
	inc di		; di has been decreased past the end -> add 1
	cli
	mov sp,di
	sti
	sub bp,ax	; adjust bp too

	mov bx,bp	; and walk through all stack frames
@@nextbp:
	mov cx,[bx]
	cmp cx,0
	je short @@bpsdone

	sub cx,ax
	mov [bx],cx
	mov bx,cx
	jmp @@nextbp


@@bpsdone:
	; now everything should be fine!

	mov bx,ax
	neg bx
	shr bx,4	; in paragraphs
	mov newsize,bx
	mov ah,4ah	; resize block
			; es is correct already
	mov es,_psp
;	int 21h

	ret
reducesize endp


;----BEGIN Interrupt handlers---------------------------------------------

; This one handles OP reg,mem; OP reg8,mem8; OP mem,reg; OP mem8,reg8
; where OP is one of MOV,ADD,OR,ADC,SBB,AND,SUB,XOR,CMP
; the OP is determined by the interrupt number, the actual kind of operation
; from the postbyte
threebyteop proc
	intro
	adjustip 1

	xor ecx,ecx	; no relative offset to be added - just so everything
			; can be reused by fourbyteop and movimm

; find out what the register operand is
decoderegister:
	mov bx,es:[bx-1]
	; bl contains the int number
	; bh contains the instruction postfix:
	;	bh = [1 D  1 S  3 Memop  3 Regop]
	;	D: 0=write, 1=read  i.e.  0=mem,reg  1==reg,mem
	;	S: 0=byte,  1=word/dword

	xor dx,dx
	mov dl,bh
	and dl,7		; now dl=RRR
	shl dx,2		; actual register value is oldeax[-4*dl]
	mov si,dx
	neg si
	mov edx,oldeax[si]	; edx is the value to be loaded
				; into the memory if writing

; find the memory operand
decodememoryoperand:
	xor ax,ax
	mov al,bh
	and al,38h		; now al=8*RRR -> al
	shr ax,1		; i.e. actual value is in oldeax[-al/2]
	mov di,ax		; now oldeax[di] is the memory operand
	neg di
	mov edi,oldeax[di]
	add edi,ecx		; now edi has the memory address

	; find the memory size	; 1=byte, 2=word, 4=dword
	mov cx,1
	test bh,40h		; byte op doesn't care about a memory size prefix
	jz short @@itsabyteop
	test prefix,msprf
	setnz cl
	inc cx
	shl cx,1
@@itsabyteop:			; it's in cx

	voodoogetmem edi,cx ; and eax gets the memory content

	; make sure that the source is in eax and dest is edx
	test bh,80h
	jnz short @@itsaread
	xchg eax,edx

	; now we want to determine the real CPU opcodes which execute the right
	; instruction, then modify the code to execute it
@@itsaread:
	cmp bl,0a0h
	je short @@domov	; MOV opcode doesn't follow the same rules
	cmp bl,0a9h
	je short @@domov
	cmp bl,0beh
	je short @@domov
	; it's one of the op1 instructions
	shl bl,3
	jmp short @@dotheop
@@domov:
	mov bl,88h

@@dotheop:
	push bx
	shr bh,6
	or bh,bl	; copy bits 6+7 from bh into bits 0+1 and combine with opcode
			; now bl has the real opcode which does the right thing

	mov bl,90h	; do we need a memory size prefix? otherwise NOP
	test prefix,msprf
	jz short @@nodword
	mov bl,66h
@@nodword:
	mov word ptr _text:@@theop,bx	; store the opcode
	pop bx

	realinstruction <or edx,eax>,@@theop

		; need to store the result
	test bh,80h
	jz short @@writetomemory

@@writetoregister:
	sizemultiplex cx,@@writeregb,@@writeregw,@@writeregd
@@writeregb:
	mov oldeax[si],dl
	jmp short @@writedone
@@writeregw:
	mov oldeax[si],dx
	jmp short @@writedone
@@writeregd:
	mov oldeax[si],edx
	jmp short @@writedone

@@writetomemory:
	voodoosetmem edi,edx,cx

@@writedone:
	; we're done!


	extro
threebyteop endp

; same as threebyteop, but it has an 8 byte offset to be added to the memory operand
fourbyteop proc
	intro
	adjustip 2

	movsx ecx,byte ptr es:[bx+1]
	jmp decoderegister
fourbyteop endp

; mov mem,imm8 or mov mem,imm depending on postbyte
movimm proc
	intro

	; how many postbytes? 1+memory size, i.e. 3 if word and 5 if dword
	mov ax,1
	test byte ptr es:[bx+1],40h
	jz short @@knowit
	mov ax,3
	test prefix,msprf
	jz short @@knowit
	mov ax,5
@@knowit:
	adjustip ax

	xor ecx,ecx
	mov edx,es:[bx+1]	; the size doesn't really matter when reading
	mov bx,es:[bx-1]
	and bx,not 80h		; to make sure we write to memory, not a register!
	jmp decodememoryoperand

	extro
movimm endp

xlatint proc		; al := [ebx+al]
	intro

	movzx eax,byte ptr oldeax
	add eax,oldebx
	voodoogetmem eax,1
	mov oldeax,al

;	adjustip 0	; xlat is two-byte
	extro
xlatint endp

stringop proc
	;; Usage of registers:
	;; eax: Memory being shuffled around; zero flag
	;; ebx: temp
	;; ecx: Loop count
	;; edx: temp
	;; esi: Size of operation (1, 2, 4)
	;; edi: +esi or -esi, depending on the direction
	intro

	mov ecx,oldecx
	test prefix,repzprf or repnzprf
	jz short @@okdotheloop		; no rep prefix -> ignore ecx

	or ecx,ecx			; check if at least one iteration is necessary
	jnz short @@okdotheloop
	jmp @@sopbreak

@@okdotheloop:

		; self-modifying code; sets the jump to the right handler
	movzx di,byte ptr es:[bx-1]	; int number: determines which string-op
	sub di,0b2h
	shl di,1
	mov ax,cs:@@soplist[di]
	sub ax,offset @@nextiteration+3
	mov word ptr cs:[@@nextiteration+1],ax

	mov edx,1
	test di,2	; opnum is odd for word/dword and even for byte ops
	jz short @@isbyteop

	test prefix,msprf
	setnz dl
	inc edx
	shl edx,1

@@isbyteop:
	mov esi,edx	; and si=mem size (2 or 4) depending on 66 prefix
	xor ebx,ebx
	test oldflags,dirflag
	setz bl
	shl ebx,1
	dec ebx
	imul ebx,esi
	mov edi,ebx	; now edi=memsize if no dir flag and -memsize if dir flag

	xor eax,eax	; ah = flags with zero flag clear

@@nextiteration:
	db 0e9h,0,0	; jump will be modified according to which operation we have

@@sopcheck:		; check for repeat prefixes, load flags into ah first!
	test prefix,repzprf
	jz short @@mayberepnz
	dec ecx
	lahf			; restore the zero flag
	jnz short @@sopbreak	; it's repz, so break if zf wasn't set
	jecxz short @@sopbreak	; or break if ecx is zero
	jmp short @@nextiteration

@@mayberepnz:
	test prefix,repnzprf
	jz short @@sopbreak
	dec ecx
	lahf
	jz short @@sopbreak
	jecxz short @@sopbreak
	jmp short @@nextiteration

@@sopcheckrep:		; check only for any repeat prefix - not specific z/nz
	test prefix,repzprf or repnzprf
	jz short @@sopbreak
	dec ecx
	jecxz short @@sopbreak
	jmp short @@nextiteration

@@sopbreak:
	mov oldecx,ecx
	extro


@@movs:
	voodoogetmem <dword ptr oldesi>,si

	voodoosetmem <dword ptr oldedi>,eax,si

	add oldesi,edi
	add oldedi,edi
	jmp @@sopcheckrep


@@cmps:
	voodoogetmem <dword ptr oldesi>,si
	mov ebx,eax

	voodoogetmem <dword ptr oldedi>,si

	add oldesi,edi
	add oldedi,edi

@@docmps:
	push oldflags
	popf
	sizemultiplex si,@@cmpsb,@@cmpsw,@@cmpsd
@@cmpsb:
	cmp bl,al
@@cmpsdone:
	pushf
	pop oldflags
	sahf
	jmp @@sopcheck
@@cmpsw:
	cmp bx,ax
	jmp @@cmpsdone
@@cmpsd:
	cmp edx,eax
	jmp @@cmpsdone


@@illegal:	; it's not a string op, so segfault
	push 17
	call segfault


@@stos:
	voodoosetmem <dword ptr oldedi>,<dword ptr oldeax>,si

	add oldedi,edi
	jmp @@sopcheckrep


@@lods:
	voodoogetmem <dword ptr oldesi>,si
	add oldesi,edi

	sizemultiplex si,@@lodsb,@@lodsw,@@lodsd
@@lodsb:
	mov oldeax,al
	jmp @@sopcheckrep
@@lodsw:
	mov oldeax,ax
	jmp @@sopcheckrep
@@lodsd:
	mov oldeax,eax
	jmp @@sopcheckrep


@@scas:
	voodoogetmem <dword ptr oldesi>,si
	add oldesi,edi

	mov ebx,oldeax
	jmp @@docmps


@@soplist dw offset @@movs,offset @@movs
	dw offset @@cmps,offset @@cmps
	dw offset @@illegal,offset @@illegal
	dw offset @@stos,offset @@stos
	dw offset @@lods,offset @@lods
	dw offset @@scas,offset @@scas
stringop endp


; some special cases that only occur once and don't warrant general support
; of the instruction
special proc
	intro

	mov al,es:[bx]

	dec al
	jo short @@testeaxw	; al was 0

	; undefined special case.  Segfault
	push 12
	call segfault

@@specialdone:		; put it in front so the jmps to here know how far it is
	adjustip ax	; depends on special.
	extro

@@testeaxw:
	voodoogetmem <dword ptr oldeax>,2

	realinstruction <test ax,es:[bx+1]>,@@theop

	mov ax,3	; how many parms follow the int
	jmp @@specialdone

special endp

end
